#!/usr/bin/env python3

#import scipy

import sys
import time
import board
import busio
import adafruit_lidarlite
from micropython import const
import RPi.GPIO as GPIO

#from threading import Thread
import multiprocessing

import led_control_funs as led

import datetime
from datetime import date, timedelta
import shutil

import ssh_mySQL_utils as sql
import mySQL_utils as localSQL

import dateTime_utils as dtu

#only new line in r7 is to use the updated version of get_Node_Info_r2 (versus original get_Node_Info)
import get_Node_Info_r2 as gni   #local module that gets: camNode_port, sGinName, GinStandNum, GinStandPositionNum for this node

import log_errors as log

#need to add modules to path so this script can access them
#sys.path.append(r'/home/pi/pides_ops/')

sMsg = "started lidarlite_test_r8.py"
log.write_debug_log(sMsg)



camNode_port, sGinName, GinStandNum, GinStandPositionNum = gni.get_Node_Info()

today = date.today()

# dd/mm/YY
sDate = today.strftime("%d/%m/%Y")
print("sDate =", sDate)

from datetime import datetime, timedelta
now = datetime.now()

sTime = now.strftime("%H:%M:%S") 

print(sTime)
print()


led_pin = 20  #GPIO20 PIN 38
reset_pin = 6 #GPIO6 for resetting Lidar


#Suppress warnings
GPIO.setwarnings(False)
#Use GPIO pin numberings
GPIO.setmode(GPIO.BCM)
#Set LED pin as output
GPIO.setup(led_pin, GPIO.OUT)
GPIO.setup(reset_pin, GPIO.IN)


#ONCE reset_pin has been enabled as output, i2c stops working!!!
#reset Lidar
#GPIO.output(led_pin, GPIO.LOW)
#time.sleep(1)
#GPIO.output(led_pin, GPIO.HIGH)


#get free disk space
def get_free_diskSpace():
    # Path
    path = "/home/pi/"
      
    # Get the disk usage statistics
    # about the given path
    stat = float(shutil.disk_usage(path)[2])
    
    diskFree_GB = int(stat/1000/1000/1000) - 15  # 15GB used elsewhere that isn't avail.  disk-swap etc.
    

    return diskFree_GB

diskFree_GB = get_free_diskSpace()
# Print disk usage statistics
print("Disk free: ", diskFree_GB)
print()


led.init_LED()


print("Startup Flash")
led.startup_flash()


 
 
_REG_ACQ_COMMAND = const(0x00)
_CMD_RESET = const(0x00)
_CMD_DISTANCENOBIAS = const(0x03)
_CMD_DISTANCEWITHBIAS = const(0x04)
_REG_DIST_MEAS_V3 = const(0x8F)
_NUM_DIST_BYTES = 2  # How many bytes is the returned distance measurement?
_REG_SIGNAL_STRENGTH = const(0x0E)

"""Status Registers"""
# v3
STATUS_BUSY = 0x01
STATUS_REF_OVERFLOW = 0x02
STATUS_SIGNAL_OVERFLOW = 0x04
STATUS_NO_PEAK = 0x08
STATUS_SECOND_RETURN = 0x10
STATUS_HEALTHY = 0x20
STATUS_SYS_ERROR = 0x40

class lidarlite_ext(adafruit_lidarlite.LIDARLite):
    def hello(this):
        print("hello from Lidarlite_ext")

    def read_distance_v3(self, bias=False):
        #print("ext distance")
        """Perform a distance reading with or without 'bias'. It's recommended
        to take a bias measurement every 100 non-bias readings (they're slower)"""
        
        try:
            if bias:
                self._write_reg(_REG_ACQ_COMMAND, _CMD_DISTANCEWITHBIAS)
            else:
                self._write_reg(_REG_ACQ_COMMAND, _CMD_DISTANCENOBIAS)
            dist = self._read_reg(_REG_DIST_MEAS_V3, _NUM_DIST_BYTES)
            if self._status & (STATUS_NO_PEAK | STATUS_SECOND_RETURN):
                if self._status & STATUS_NO_PEAK:
                    #raise RuntimeError("Measurement failure STATUS_NO_PEAK")
                    print("Measurement failure STATUS_NO_PEAK")
                if self._status & STATUS_SECOND_RETURN:
                    #raise RuntimeError("Measurement failure STATUS_NO_PEAK")
                    print("Measurement failure STATUS_NO_PEAK")
            
                #raise RuntimeError("Some other runtime error")
                return 0
            if (self._status & STATUS_SYS_ERROR) or (not self._status & STATUS_HEALTHY):
                print("System failure, line 52")
                raise RuntimeError("System failure")
            
            distance = dist[0] << 8 | dist[1]
        except Exception:
            distance = -99
        except:
            #pass
            distance = -999
        
        return distance

    @property
    def distance(self):  # pylint: disable=R1710
        """The measured distance in cm. Will take a bias reading every 100 calls"""
        self._bias_count -= 1

        if self._bias_count < 0:
            self._bias_count = 100  # every 100 reads, check bias
            
        return self.read_distance_v3(self._bias_count <= 0)

        # If no sensor type has been identified, return a negative distance as an error
        return -1.0

    @property
    def signal_strength(self):
        """Reads the signal strength of the last measurement"""
        return self._read_reg(_REG_SIGNAL_STRENGTH, 1)[0]



# code to populate Lidar dbase
sDataBase = "camera_nodes"
sServer = "cotton@45.56.121.38"
ssh_port = 22
#sPort = str(port)
#sPortArg = "-p " + sPort

sUser = "ginuser"
sPswd = "Hello2018"


tableName = "Lidar"

def write_localSQL(insert_query):
    print()
    print("thread writing to localSQL")
    print()
    cnx = localSQL.sql_connect()
    
    localSQL.sql_insert(cnx, insert_query)

    localSQL.sql_closeConnection(cnx)
    
    
def parallel_write_event_localSQL(process_list, dist, rssi, numBreaks_detected, sDate, sTime, epochTime):
    
    #nodId on cam-node doesn't mean anything so set to 0.  on Server, it will be the id for the Lidar table on cam-node
    #similarly for 'Active' and 'port'
    insert_query = 'INSERT INTO Lidar (BreakDetected, RSSI, Distance, Date, Time, Epoch) VALUES (%s, %s, %s, "%s", "%s", %s)'%( str(numBreaks_detected),str(rssi),str(dist), sDate, sTime, str(epochTime) )
    
    print()
    print("parallel write query: ", insert_query)
    print()
    
    #args=(insert_query,)  (either use tuple args i.e. (query, )  "don't miss the ','!!!  or pass string as list (as below)
    #t1 = Thread(target=write_localSQL, args=[insert_query])   #when passing string as arg, use [] to indicate list or you'll get error
    #t1.daemon = True   #set as a deamon thread so that thread is killed if/when program is killed
    #t1.start()

    #go through process_list and kill any unterminated processes
    for proc in process_list:
        if(proc.is_alive()):
            proc.terminate()
        else:
            #remove from list
            process_list.remove(proc)

    process = multiprocessing.Process(target=write_localSQL, args=[insert_query])   #when passing string as arg, use [] to indicate list or you'll get error
    process.start()
    
    process_list.append(process)
    
    return process_list


def write_event_localSQL(dist, rssi, numBreaks_detected, sDate, sTime, epochTime):
    
   
    cnx = localSQL.sql_connect()
    
    #nodId on cam-node doesn't mean anything so set to 0.  on Server, it will be the id for the Lidar table on cam-node
    #similarly for 'Active' and 'port'
    insert_query = 'INSERT INTO Lidar (BreakDetected, RSSI, Distance, Date, Time, Epoch) VALUES (%s, %s, %s, "%s", "%s", %s)'%( str(numBreaks_detected),str(rssi),str(dist), sDate, sTime, str(epochTime) )
    print(insert_query)
    localSQL.sql_insert(cnx, insert_query)

    localSQL.sql_closeConnection(cnx)
    

#write data to local Lidar dbase
def write_event_remoteSQL(dist, rssi, numBreaks_detected, sDate, sTime):
    print("writing to sql dbase")
    
    sQuery = 'INSERT INTO Lidar (port, GinName, GinStandNum, BreakDetected, Distance, RSSI, Date, Time) VALUES (%s, "%s", %s, %s, %s, %s, "%s", "%s");'%(str(camNode_port), sGinName, str(GinStandNum), str(numBreaks_detected), str(int(dist)), str(int(rssi)), (sDate), (sTime)) 
    #print(sQuery)

    try:
        outs = sql.mySQL_over_SSH(sServer, ssh_port, sUser, sPswd, sDataBase, sQuery)
        #outs, ers = mySQL_over_SSH(sServer, port, sUser, sPswd, sDataBase, sQuery)

        print(outs)
    except:
        sMsg="error writing data to SQL on Cloud; line 212 lidarlite_test_r8.py"
        print(sMsg)
        log.write_debug_log(sMsg)
        


def reset_lidar(sensorExt):
    
    try:
        sensor.reset()
    except:
        sMsg="error reseting Lidar"
        print(sMsg)
        log.write_debug_log(sMsg)


def read_lidar(sensorExt):
    dist = -9999
    rssi = -9999

    try:
        dist = int(sensorExt.distance)
        rssi = sensorExt.signal_strength
        print("DIST: ", dist)
        print("RSSI: ", rssi)
        
    except:
        sMsg="error reading Lidar"
        print(sMsg)
        log.write_debug_log(sMsg)


    return dist, rssi


def lowPassFilter(dist, lpf_dist):
    
    lpf_dist = (lpf_dist + dist) / 2.0
    
    return lpf_dist


#END OF FUNCTIONS:



sensorExt = any
MAX_NUM_LOOPS = 4
# try MAX_NUM_LOOPS times to setup (connect to) lidar unit
for i in range(MAX_NUM_LOOPS): 
    # Setup Lidar unit:
    try:
        # Create library object using our Bus I2C port
        i2c = busio.I2C(board.SCL, board.SDA)
        
 #NOT NEEDED?       sensor = adafruit_lidarlite.LIDARLite(i2c, address=0x62)
 #ORIG       sensorExt = lidarlite_ext(i2c, address=0x62)

        sensorExt = lidarlite_ext(i2c, configuration=1, address=0x62)   #config=1 is CONFIG_SHORTFAST
        #sensorExt = lidarlite_ext(i2c, configuration=4, address=0x62)  #config=4 is CONFIG_HIGHSENSITIVE
       

    except:
        sMsg="cam-node: " + str(camNode_port) + " | error obtaining i2C bus, line 266 lidarlite_test_r8.py"
        print(sMsg)
        log.write_debug_log(sMsg)

        time.sleep(10)

        if(i>=MAX_NUM_LOOPS-1):
            sMsg="giving up & exiting: cam-node: " + str(camNode_port) + " | error obtaining i2C bus, line 266 lidarlite_test_r6.py"
            print(sMsg)
            log.write_debug_log(sMsg)

            sys.exit()
    else:  #executes if successful, so exit loop
        break


#device_id = sensor._read_register_byte(adafruit_lidarlite._REG_WHOAMI)
#print("device_id: ", device_id)

#if(device_id != 0x33):
#    raise RuntimeError("Failed to find LIDAR-LITE!!")


# Optionally, we can pass in a hardware reset pin, or custom config
# import digitalio
# reset = digitalio.DigitalInOut(board.D5)
#sensor = adafruit_lidarlite.LIDARLite(i2c, reset_pin=reset,
#    configuration=adafruit_lidarlite.CONFIG_MAXRANGE)
#CONFIG_SHORTFAST, CONFIG_LOWSENSITIVE, CONFIG_HIGHSENSITIVE, 

# If you want to reset, you can do so, note that it can take 10-20 seconds
# for the data to 'normalize' after a reset (and this isnt documented at all)
print("lidarlite reset; wait 10-20 seconds")
reset_lidar(sensorExt)

print("start:")
print()

sleep_ms = 0.1

#production threahold
short_dist_threshold = 220  #243cm == 8ft  so back off 12" == 30.5cm
long_dist_threshold = 285


reportTime_secs = 150  #test report time; production reduce to 60 (1 every minute; results in disk storage of 10MB per season, so no worries)

#start delta-timer
start_time = datetime.now()

breakEvent_startTime = datetime.now()
min_break_reportTime_secs = 60  #(don't record more than once per minute) (so once we register a beam break, wait until 60 seconds before registering another one ?)


break_detection_threshold = 1  # of breaks that must be recorded in order to trigger a true beam-break
break_detected = 0

MAX_NUM_LOOPS_CHECK_GIN_NAME = 600
led_state = 0
indx = 0

lpf_dist = 270.0  #init to about normal dist


#list of spawned multiprocessing processes
all_processes = []

try:
    while True:

        indx += 1 # current loop iteration
        
        # periodically, stop and check for new ginName  (ONLY CHECK LOCAL DATA-BASE SO AS TO NOT INTERFERE WITH INTRUSION-DETECTION)
        # LOCAL DATABASE WILL GET UPDATED IN SEPARATE SYSTEMD SERVICE
        if(indx >= MAX_NUM_LOOPS_CHECK_GIN_NAME):
            indx = 0  #reset cntr
            #see if need to go out to network and get real name, if still using "na" as a gin-name we do
            if(sGinName == "na"):
                camNode_port, sGinName, GinStandNum, GinStandPositionNum = gni.get_Node_Info(get_only_local_nodeInfo = True)
                sMsg="in while loop, lidarlite_test7, with sGinName of na, line# 331"
                print(sMsg)
                log.write_debug_log(sMsg)


        
        #toggle_led(led_state)
        print("toggle LED")
        led_state = led.toggle_led(led_state)

        
        print()
        print()
        
        today = date.today()

        # dd/mm/YY
        sDate = today.strftime("%d/%m/%Y")
        print("date: ", sDate)
        
        sTime = time.strftime("%H:%M:%S", time.localtime())
     
        print("Current Time is :", sTime)
        print()

        #dist = int(sensorExt.distance)
        #rssi = sensorExt.signal_strength
        dist, rssi = read_lidar(sensorExt) 
        print("DIST: ", dist)
        print("RSSI: ", rssi)
        
        
        # check if beam broken
        # Could we get an error that occurs when beam is broken, thus not registering it as a beam break. (testable)
        if(dist > 0):  #verify valid reading
            
            #ignore any zero readings as they are bad readings, so put in after test for dist>0
            lpf_dist = lowPassFilter(dist, lpf_dist)
            print()
            print("lpf_dist: ", lpf_dist)
            print()
            
            #check if beam broken
            if(dist > 0):  #weed out errors -999's
                if((dist < short_dist_threshold) or (dist > long_dist_threshold)):
                    break_detected += 1 # running total
                    
                    #get new start-time
                    cur_time = datetime.now() # keep track of time when a beam break occurs

                    
                    #
                    #breakEvent_startTime = datetime.now()
                    # Get different between two datetime as timedelta object.
                    break_timedelta_obj = (cur_time - breakEvent_startTime) # how long from first execution of program to when beam break occurs
            
                    break_deltaTime_secs = break_timedelta_obj.total_seconds()
                
               
           # else:  #beam not broken
        #record beam-break event
        print("beam-break count: ", break_detected)
            
        
        #check if time for keep alive update
        cur_time = datetime.now()
        
        # Get different between two datetime as timedelta object.
        timedelta_obj = (cur_time - start_time)
        
        deltaTime_secs = timedelta_obj.total_seconds()
        deltaTime_mins = timedelta_obj.total_seconds() / 60
        
        print(deltaTime_secs)
        
        if(deltaTime_secs > reportTime_secs):
            #get new start-time
            start_time = datetime.now()
            
            print("report time: ")
            
            beamBroken = False
            if(break_detected > break_detection_threshold):
                beamBroken = True
                print("Beam Broken Event")
            
            epochTime_secs = dtu.get_currentEpochTime_secs()
            
            avg_dist = int(lpf_dist + 0.5)  #round to nearest int
            print()
            print("avg dist: ", avg_dist)
            print()
            
            #write to dbase keep-alive
            
            all_processes = parallel_write_event_localSQL(all_processes, avg_dist, rssi, break_detected, sDate, sTime, epochTime_secs)
            
            #reset break count
            break_detected = 0
        

        #sleep pause is based on dist; goes slower the farther away it is
        sleep_s = float(dist) / 500.0

        print("sleep_s: ", sleep_s)

        #except RuntimeError as e:
            # If we get a reading error, just print it and keep truckin'
            #print(e)
            
        if(sleep_s < 0.05):
            sleep_s = 0.05
            
        time.sleep(sleep_s)  # you can remove this for ultra-fast measurements!
except:
    print("error in main line530, lidarlite_r8.py")